import os
import tempfile
from collections import defaultdict
import logging
from distutils.dir_util import copy_tree
from functools import partial
from shutil import copyfile
import hashlib
import tarfile
import re
from zipfile import ZipFile
from git import Repo

from kapitan.errors import GitSubdirNotFoundError
from kapitan.utils import make_request

logger = logging.getLogger(__name__)

# this is used just to change the root directory of save_dir in testing
DEPENDENCY_OUTPUT_CONFIG = {}


def fetch_dependencies(target_objs, pool):
    """
    parses through the dependencies parameters in target_objs and fetches them
    all dependencies are first fetched into tmp dir, after which they are copied to their respective output_path.
    dependencies whose output_path already exists are skipped
    """
    temp_dir = tempfile.mkdtemp()
    # there could be multiple dependency items per source_uri due to reclass inheritance or
    # other user requirements. So create a mapping from source_uri to a set of dependencies with
    # that source_uri
    git_deps = defaultdict(list)
    http_deps = defaultdict(list)

    # this dict is to make sure no duplicated output_paths exist per source
    deps_output_paths = defaultdict(set)
    for target_obj in target_objs:
        try:
            dependencies = target_obj["dependencies"]
            for item in dependencies:
                dependency_type = item["type"]
                source_uri = item["source"]
                if "root_dir" in DEPENDENCY_OUTPUT_CONFIG:
                    # should only be used in test environment
                    item["output_path"] = os.path.join(DEPENDENCY_OUTPUT_CONFIG["root_dir"], item["output_path"])
                output_path = item["output_path"]
                if os.path.exists(os.path.abspath(output_path)) and not item.get('unpack', False):
                    # if unpack: True, then allow the user to specify the parent directory as output_path where the
                    # files will be extracted to
                    logger.info("Dependency {} : already exists. Ignoring".format(output_path))
                    continue

                if output_path in deps_output_paths[source_uri]:
                    # if the output_path is duplicated for the same source_uri
                    continue
                else:
                    deps_output_paths[source_uri].add(output_path)

                if dependency_type == "git":
                    git_deps[source_uri].append(item)
                elif dependency_type in ("http", "https"):
                    http_deps[source_uri].append(item)

        except KeyError:
            continue

    git_worker = partial(fetch_git_dependency, save_dir=temp_dir)
    http_worker = partial(fetch_http_dependency, save_dir=temp_dir)
    [p.get() for p in pool.imap_unordered(http_worker, http_deps.items()) if p]
    [p.get() for p in pool.imap_unordered(git_worker, git_deps.items()) if p]


def fetch_git_dependency(dep_mapping, save_dir):
    """
    fetches a git repository at source into save_dir, and copy the repository into output_path
    ref is used to checkout if exists
    only subdir is copied into output_path if specified
    """
    source, deps = dep_mapping
    fetch_git_source(source, save_dir)
    for dep in deps:
        repo_path = os.path.join(save_dir, os.path.basename(source))
        repo = Repo(repo_path)
        output_path = dep["output_path"]
        copy_src_path = repo_path
        if 'ref' in dep:
            ref = dep['ref']
            repo.git.checkout(ref)
        else:
            repo = Repo(repo_path)
            repo.git.checkout("master")  # default ref

        if 'subdir' in dep:
            sub_dir = dep['subdir']
            full_subdir = os.path.join(repo_path, sub_dir)
            if os.path.isdir(full_subdir):
                copy_src_path = full_subdir
            else:
                raise GitSubdirNotFoundError("Dependency {} : subdir {} not found in repo".format(source, sub_dir))

        if not os.path.exists(os.path.abspath(output_path)):
            copy_tree(copy_src_path, output_path)
            logger.info("Dependency {} : saved to {}".format(source, output_path))


def fetch_git_source(source, save_dir):
    """clones a git repository at source and saves into save_dir"""
    logger.info("Dependency {} : fetching now".format(source))
    Repo.clone_from(source, os.path.join(save_dir, os.path.basename(source)))
    logger.info('Dependency {} : successfully fetched'.format(source))


def fetch_http_dependency(dep_mapping, save_dir):
    """
    fetches a http[s] file at source and saves into save_dir, after which it is copied into
    the output_path stored in dep_mapping
    """
    source, deps = dep_mapping
    content_type = fetch_http_source(source, save_dir)
    path_hash = hashlib.sha256(os.path.dirname(source).encode()).hexdigest()[:8]
    copy_src_path = os.path.join(save_dir, path_hash + os.path.basename(source))
    for dep in deps:
        output_path = dep["output_path"]
        if dep.get('unpack', False):
            # ensure that the directory we are extracting to exists
            os.makedirs(output_path, exist_ok=True)
            is_unpacked = False
            if content_type == 'application/x-tar':
                tar = tarfile.open(copy_src_path)
                tar.extractall(path=output_path)
                tar.close()
                is_unpacked = True
            elif content_type == 'application/zip':
                zfile = ZipFile(copy_src_path)
                zfile.extractall(output_path)
                zfile.close()
                is_unpacked = True
            elif content_type in ['application/octet-stream', 'application/x-gzip']:
                if re.search(r'(\.tar\.gz|\.tgz)$', source):
                    tar = tarfile.open(copy_src_path)
                    tar.extractall(path=output_path)
                    tar.close()
                    is_unpacked = True
            if is_unpacked:
                logger.info("Dependency {} : extracted to {}".format(source, output_path))
            else:
                logger.info("Dependency {} : Content-Type {} is not supported for unpack. Ignoring save".
                            format(source, content_type))
        else:
            # we are downloading a single file
            parent_dir = os.path.dirname(output_path)
            if parent_dir != '':
                os.makedirs(parent_dir, exist_ok=True)
            copyfile(copy_src_path, output_path)
            logger.info("Dependency {} : saved to {}".format(source, output_path))


def fetch_http_source(source, save_dir):
    """downloads a http[s] file from source and saves into save_dir"""
    logger.info("Dependency {} : fetching now".format(source))
    content, content_type = make_request(source)
    logger.info("Dependency {} : successfully fetched".format(source))
    if content is not None:
        basename = os.path.basename(source)
        # to avoid collisions between basename(source)
        path_hash = hashlib.sha256(os.path.dirname(source).encode()).hexdigest()[:8]
        full_save_path = os.path.join(save_dir,  path_hash + basename)
        with open(full_save_path, 'wb') as f:
            f.write(content)
        return content_type
    else:
        logger.warning("Dependency {} : failed to fetch".format(source))
        return None



# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"cli tests"

import base64
import contextlib
import io
import os
import tempfile
import shutil
import subprocess
import sys
import unittest

from kapitan.cli import main

REFS_PATH = tempfile.mkdtemp()

# set GNUPGHOME if only running this test
# otherwise it will reuse the value from test_gpg.py
if os.environ.get("GNUPGHOME", None) is None:
    GNUPGHOME = tempfile.mkdtemp()
    os.environ["GNUPGHOME"] = GNUPGHOME


class CliFuncsTest(unittest.TestCase):
    def setUp(self):
        example_key = 'examples/kubernetes/refs/example@kapitan.dev.key'
        example_key = os.path.join(os.getcwd(), example_key)
        example_key_ownertrust = tempfile.mktemp()

        # always trust this key - for testing only!
        with open(example_key_ownertrust, "w") as fp:
            fp.write("D9234C61F58BEB3ED8552A57E28DC07A3CBFAE7C:6\n")

        subprocess.run(["gpg", "--import", example_key])
        subprocess.run(["gpg", "--import-ownertrust", example_key_ownertrust])
        os.remove(example_key_ownertrust)

    def test_cli_secret_write_reveal_gpg(self):
        """
        run $ kapitan refs --write gpg:test_secret
        and $ kapitan refs --reveal
        with example@kapitan.dev recipient
        """
        test_secret_content = "I am a secret!"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = ["kapitan", "refs", "--write", "gpg:test_secret",
                    "-f", test_secret_file,
                    "--refs-path", REFS_PATH,
                    "--recipients", "example@kapitan.dev"]
        main()

        test_tag_content = "revealing: ?{gpg:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal",
                    "-f", test_tag_file,
                    "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("revealing: {}".format(test_secret_content),
                         stdout.getvalue())

        os.remove(test_tag_file)

    def test_cli_secret_base64_write_reveal_gpg(self):
        """
        run $ kapitan refs --write gpg:test_secretb64 --base64
        and $ kapitan refs --reveal
        with example@kapitan.dev recipient
        """
        test_secret_content = "I am another secret!"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = ["kapitan", "refs", "--write", "gpg:test_secretb64",
                    "-f", test_secret_file, "--base64",
                    "--refs-path", REFS_PATH,
                    "--recipients", "example@kapitan.dev"]
        main()

        test_tag_content = "?{gpg:test_secretb64}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal",
                    "-f", test_tag_file,
                    "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        stdout_base64 = base64.b64decode(stdout.getvalue()).decode()
        self.assertEqual(test_secret_content, stdout_base64)

        os.remove(test_tag_file)

    def test_cli_secret_validate_targets(self):
        """
        run $ kapitan refs --validate-targets
        expect 0 (success) exit status code
        """
        with self.assertRaises(SystemExit) as cm:
            sys.argv = ["kapitan", "refs", "--validate-targets",
                        "--refs-path", "examples/kubernetes/refs/targets/",
                        "--inventory-path", "examples/kubernetes/inventory/"]
            main()
        self.assertEqual(cm.exception.code, 0)

    def test_cli_secret_write_reveal_gkms(self):
        """
        run $ kapitan refs --write gkms:test_secret
        and $ kapitan refs --reveal
        using mock KMS key
        """
        test_secret_content = "mock"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = ["kapitan", "refs", "--write", "gkms:test_secret",
                    "-f", test_secret_file,
                    "--refs-path", REFS_PATH,
                    "--key", "mock"]
        main()

        test_tag_content = "revealing: ?{gkms:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal",
                    "-f", test_tag_file,
                    "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("revealing: {}".format(test_secret_content),
                         stdout.getvalue())

        os.remove(test_tag_file)

    def test_cli_secret_write_reveal_awskms(self):
        """
        run $ kapitan refs --write awskms:test_secret
        and $ kapitan refs --reveal
        using mock KMS key
        """
        test_secret_content = "mock"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = ["kapitan", "refs", "--write", "awskms:test_secret",
                    "-f", test_secret_file,
                    "--refs-path", REFS_PATH,
                    "--key", "mock"]
        main()

        test_tag_content = "revealing: ?{awskms:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal",
                    "-f", test_tag_file,
                    "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("revealing: {}".format(test_secret_content),
                         stdout.getvalue())

        os.remove(test_tag_file)

    def test_cli_secret_write_plain_ref(self):
        """
        run $ kapitan refs --write plain:test_secret
        and $ kapitan refs --reveal -f sometest_file
        """
        test_secret_content = "secret_value!"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = ["kapitan", "refs", "--write", "plain:test_secret",
                    "-f", test_secret_file,
                    "--refs-path", REFS_PATH]
        main()

        test_tag_content = "revealing: ?{plain:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal",
                    "-f", test_tag_file,
                    "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("revealing: {}".format(test_secret_content),
                         stdout.getvalue())

        os.remove(test_tag_file)

    def test_cli_secret_write_base64_ref(self):
        """
        run $ kapitan refs --write base64:test_secret
        and $ kapitan refs --reveal -f sometest_file
        """
        test_secret_content = "secret_value!"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = ["kapitan", "refs", "--write", "base64:test_secret",
                    "-f", test_secret_file,
                    "--refs-path", REFS_PATH]
        main()

        test_tag_content = "revealing: ?{base64:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal",
                    "-f", test_tag_file,
                    "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("revealing: {}".format(test_secret_content),
                         stdout.getvalue())

        os.remove(test_tag_file)

    def test_cli_secret_write_base64_encoded_ref(self):
        """
        run $ kapitan refs --write base64:test_secret --base64
        and $ kapitan refs --reveal -f sometest_file
        """
        test_secret_content = "secret_value!"
        test_secret_content_b64 = base64.b64encode(test_secret_content.encode())
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = ["kapitan", "refs", "--write", "base64:test_secret",
                    "--base64", "-f", test_secret_file,
                    "--refs-path", REFS_PATH]
        main()

        test_tag_content = "revealing: ?{base64:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal",
                    "-f", test_tag_file,
                    "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("revealing: {}".format(test_secret_content_b64.decode()),
                         stdout.getvalue())

        os.remove(test_tag_file)

    def test_cli_secret_subvar_base64_ref(self):
        """
        run $ kapitan refs --write base64:test_secret
        and $ kapitan refs --reveal -f sometest_file
        """
        test_secret_content = """
        var1:
          var2: hello
        var3:
          var4: world
        """
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = ["kapitan", "refs", "--write", "base64:test_secret_subvar",
                    "-f", test_secret_file,
                    "--refs-path", REFS_PATH]
        main()

        test_tag_content = """
        revealing1: ?{base64:test_secret_subvar@var1.var2}
        revealing2: ?{base64:test_secret_subvar@var3.var4}
        """
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal",
                    "-f", test_tag_file,
                    "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()

        expected = """
        revealing1: {}
        revealing2: {}
        """
        self.assertEqual(expected.format("hello", "world"),stdout.getvalue())
        os.remove(test_tag_file)

    def test_cli_secret_subvar_gpg(self):
        """
        run $ kapitan refs --write gpg:test_secret
        and $ kapitan refs --reveal -f sometest_file
        """
        test_secret_content = """
                var1:
                  var2: hello
                var3:
                  var4: world
                """
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        sys.argv = ["kapitan", "refs", "--write", "gpg:test_secret_subvar",
                    "-f", test_secret_file,
                    "--refs-path", REFS_PATH,
                    "--recipients", "example@kapitan.dev"]
        main()

        test_tag_content = """
                revealing1: ?{gpg:test_secret_subvar@var1.var2}
                revealing2: ?{gpg:test_secret_subvar@var3.var4}
                """
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        sys.argv = ["kapitan", "refs", "--reveal",
                    "-f", test_tag_file,
                    "--refs-path", REFS_PATH]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()

        expected = """
                revealing1: {}
                revealing2: {}
                """
        self.assertEqual(expected.format("hello", "world"), stdout.getvalue())
        os.remove(test_tag_file)

    def test_cli_searchvar(self):
        """
        run $ kapitan searchvar mysql.replicas
        """
        sys.argv = ["kapitan", "searchvar", "mysql.replicas",
                    "--inventory-path", "examples/kubernetes/inventory/"]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("examples/kubernetes/inventory/targets/minikube-mysql.yml   1\n", stdout.getvalue())

    def test_cli_inventory(self):
        """
        run $ kapitan inventory -t minikube-es -F -p cluster
        """
        sys.argv = ["kapitan", "inventory", "-t", "minikube-es", "-F", "-p", "cluster",
                    "--inventory-path", "examples/kubernetes/inventory/"]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            main()
        self.assertEqual("id: minikube\nname: minikube\ntype: minikube\nuser: minikube\n",
                         stdout.getvalue())

    def tearDown(self):
        shutil.rmtree(REFS_PATH, ignore_errors=True)
